////////////////////////////////////////////////////////
// Game Server programming
// create by Kim Dong-sung
// Update 2009.01.11
////////////////////////////////////////////////////////
#include "Search_Neighbors.h"
//#define TESTING

// coordinate system value
pt_list **world_coord = NULL;

// count value
int id_count = 0;                       // ID count
int neighbors_count = 1;                // neighbor count

// add ID point
int add_point(int id, float x, float y)
{
	point_t data;
	int i, j;               // insert coord x,y
	
	data.id = id;
	data.x = x;
	data.y = y;

	i = (int)(x / cell_width) + Origin_coord_X;
	j = (int)(y / cell_height) + Origin_coord_Y;

	printf("insert client location [%d, %d]\n", j, i);

	world_coord[j][i].push_back(data);
	return 0;
}

int find_point(int id, int coord_x, int coord_y)
{
	point_t temp;
	temp.id = id;
	pt_list::iterator pos;

	pos = find(	world_coord[coord_y][coord_x].begin(),
		world_coord[coord_y][coord_x].end(),
		temp);

	if(pos != world_coord[coord_y][coord_x].end())
		return 1;
	else
		return 0;
}

// delete ID point
int del_point(int id)
{
	point_t temp;
	temp.id = id;

	for (int i=0; i<height_part_cnt; ++i) 
	{
		for(int j=0; j<width_part_cnt; ++j)
		{
			if(find_point(id, i, j))
			{
				world_coord[j][i].remove(temp);
#ifdef TESTING
				printf("Delete ID cooredinates is %d : (%d, %d)\n", id, i,j);
#endif
				return 1;
			}
		}
	}
	return 0;
}

// delete ID point
void del_point(int id, int coord_x, int coord_y)
{
	point_t temp;
	temp.id = id;

	world_coord[coord_y][coord_x].remove(temp);
}

// point move
int move_point(int id, float distance)
{
	point_t data;
	int origin_coord_x, origin_coord_y; // id origin coord x,y
	float x,y;
	data.id = id;

	x = (float)(cos(ClientData[id].direction * 3.14/180) * distance);  
	y = (float)(-sin(ClientData[id].direction * 3.14/180) * distance);

	if(area_min.x < x && x < area_max.x &&
		area_min.y < y && y < area_max.y)
	{
		// seek origin coord
		origin_coord_x = ClientData[id].exist_cood_x;
		origin_coord_y = ClientData[id].exist_cood_y;

		ClientData[id].pos_x += x;
		ClientData[id].pos_y += y;

		// compute insert coord
		ClientData[id].exist_cood_x =
			(int)(x / cell_width) + Origin_coord_X;
		ClientData[id].exist_cood_y =
			(int)(y / cell_height) + Origin_coord_Y;

		// if input id is not existenoe
		if(ClientData[id].exist_cood_x != origin_coord_x ||
			ClientData[id].exist_cood_y != origin_coord_y)
		{
			int save = ClientData[id].situation;

			del_point(data.id, origin_coord_x,  origin_coord_y);
			add_point(data.id, ClientData[id].pos_x, ClientData[id].pos_y);

			ClientData[id].situation = save;
		}
		return 1;
	}
	else
		return 0;
}

// Ÿ ̸ 0 
int get_length(float myX, float myY, float destX, float destY, float range)
{
	float length;
	float temp;
	int result = 0;

	temp = ((myX - destX) * (myX - destX)) + ((myY - destY) * (myY - destY));
	length = sqrt(temp);

	if(length > range)
		result = -1;

	return result;
}

// print to my neighbors
void get_neighbors_data(int coord_x, int coord_y, float ox, float oy, float range, int *plist, int &cnt)
{
	pt_list::iterator pl;
	for(pl = world_coord[coord_y][coord_x].begin();
		pl != world_coord[coord_y][coord_x].end();
		++pl)
	{
		if(neighbors_count < MAX_NEIGHBORS)
		{
			if(get_length(ox, oy, pl->x, pl->y, range) == 0)
			{
#ifdef TESTING
				cout << " ["<< pl->id << "|" << pl->x << "," << pl->y << "] " << endl;
#endif
				plist[cnt++] = pl->id;
				++neighbors_count;
			}
		}
	}
}

// range list print
void range_print(int coord_x, int coord_y, float ox, float oy, float range, int replay, int *plist)
{
	// it is
	//   5 5 5 5 5 5 ...
	//   5 3 3 3 3
	//   5 3 1 1 4
	//   5 3 0 2 4
	//   5 2 2 2 4
	//   4 4 4 4 4     algorithm print result

	int dir = 0;
	int cnt = 0;

	get_neighbors_data(coord_x, coord_y, ox, oy, range, plist, cnt);

#ifdef TESTING
	printf("world coord(%d,%d)\n", coord_x, coord_y);
#endif

	for(int i=0; i < (replay<<1); ++i)
	{
		for(int j=0; j < 2; ++j)
		{
			for(int k=0; k <= i; ++k)
			{
				switch(dir)
				{
				case 0:		++coord_y;		break;
				case 1:		--coord_x;		break;
				case 2:		--coord_y;		break;
				case 3:		++coord_x;		break;
				}
				//    
				if(coord_x < 0)			coord_x = 0;
				if(coord_y < 0)			coord_y = 0;
				if(coord_x >= width_part_cnt) 
					coord_x = (int)(width_part_cnt-1);
				if(coord_y >= height_part_cnt) 
					coord_y = (int)(height_part_cnt-1);

#ifdef TESTING
				printf("world coord(%d,%d)\n", coord_x, coord_y);
#endif
				get_neighbors_data(coord_x, coord_y, ox, oy, range, plist, cnt);

				if(neighbors_count == MAX_NEIGHBORS)
					return;
			}
			(dir > 2) ? dir = 0 : ++dir;
		}
	}
}

// get neighbors
int get_neighbors(float ox, float oy, float range, int *plist)
{
	int i,j;

	// now my coord
	int my_coord_posX, my_coord_posY;

	// search coord
	int search_coord_startX, search_coord_endX;
	int search_coord_startY, search_coord_endY;
	
	// compute my coord
	my_coord_posX = 
		(int)(ox / cell_width) + Origin_coord_X;
	my_coord_posY =
		(int)(oy / cell_height) + Origin_coord_Y;

	if(my_coord_posX >= width_part_cnt)
		my_coord_posX = (int)(width_part_cnt-1);

	if(my_coord_posY >= height_part_cnt)
		my_coord_posY = (int)(height_part_cnt-1);

	// search range coord
	i = (int)(range / cell_width);
	j = (int)(range / cell_height);

	if(i < 1)               i = 1;
	if(j < 1)               j = 1;

	// comput search coord
	search_coord_startX = my_coord_posX - i;
	search_coord_endX = my_coord_posX + j;

	search_coord_startY = my_coord_posY - i;
	search_coord_endY = my_coord_posY + j;

	// search range into the global coord 
	if(search_coord_startX < 0)
		search_coord_startX = 0;
	if(search_coord_endX >= width_part_cnt)
		search_coord_endX = (int)(width_part_cnt-1);

	if(search_coord_startY < 0)
		search_coord_startY = 0;
	if(search_coord_endY >= height_part_cnt)
		search_coord_endY = (int)(height_part_cnt-1);

	// it has neighbors list print to compute info
#ifdef TESTING
	printf("my position is (%2.2f, %2.2f) \n", ox, oy);
	printf("search area cell world_width(%d ~ %d) | world_height(%d ~ %d)\n",
		search_coord_startX, search_coord_endX, search_coord_startY, search_coord_endY);
	printf("My neighbors print (max %d ids)\n", MAX_NEIGHBORS);
#endif
	neighbors_count = 0;
	range_print(my_coord_posX, my_coord_posY, ox, oy, range, (i+j)>>1, plist);

#ifdef TESTING
	printf("My(%2.2f, %2.2f) range(%2.2f) neighbors count is %d\n", ox, oy, range, neighbors_count);
#endif

	return neighbors_count;
}